/*----------------------------------------------------------------------*/
/*		hsortv.cpp		hybrid radix / merge sort						*/
/*																		*/
/*		Jeff Reid		13APR25	 10:00									*/
/*----------------------------------------------------------------------*/
/*		abbreviations for name prefixes									*/
/*																		*/
/*		a		array of ...											*/
/*		b		unsigned char (byte)									*/
/*		cb		count of bytes in ...									*/
/*		fp		FILE pointer											*/
/*		p		pointer	to ...											*/
/*		dw		unsigned int (dword)									*/
/*----------------------------------------------------------------------*/

/*		get	rid	of warnings	for	fopen */
#define	_CRT_SECURE_NO_WARNINGS	1

#include <algorithm>
#include <fstream>
#include <iomanip>
#include <iostream>
#include <vector>

#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#define	NUMREC (1024*4096)				// number of records
#define	NUMBIN 32						// number of bins
#define	SHFBIN 59						// shift factor	for	NUMBIN (256:56 128:57 64:58	32:59 16:60	8:61 4:62 2:63)

typedef	unsigned int		UI32;
typedef	unsigned int  *		PUI32;
typedef	unsigned __int64	UI64;
typedef	unsigned __int64 *	PUI64;
typedef	unsigned __int64 **	PPUI64;

#define	CMPFUNC	0						// set to 1	to use compare function

#if	CMPFUNC
typedef	bool (*PFCMP)(UI64,	UI64);
static bool	ElemCmp(UI64, UI64);
PFCMP pfCmp	= ElemCmp;
static UI64	qwCmp;
#endif

static clock_t dwTimeStart;				// clock values
static clock_t dwTimeStop;

static PUI64	pData;					// ptr to input	data
static PUI64	pSort;					// ptr to sorted data
static PUI32	pBinc;					// ptr to bin counts
static PPUI64	pBinp;					// ptr to bin ptrs
static PUI64	pBind;					// ptr to bin data
static PUI64	pDone;					// ptr to sorted bin data
static FILE	*	fpDst;					// file	desc dest file

static void	MakeBind(void);
static void	SortBind(void);
static void	MakeGroups(PUI64, UI32);
static void	MergeGroups(PUI64, PUI64, UI32);

int	main( )
{
using namespace	std;

	vector <UI64>  vData (NUMREC);		// allocate	vectors
	vector <UI32>  vBinc (NUMBIN);
	vector <PUI64> vBinp (NUMBIN);
	vector <UI64>  vBind (NUMREC * NUMBIN);
	vector <UI64>  vSort (NUMREC);

	pData =	&vData[0];
	pSort =	&vSort[0];
	pBind =	&vBind[0];
	pBinc =	&vBinc[0];
	pBinp =	&vBinp[0];

	UI64 i,	j;

//		generate data

	for(j =	0; j < NUMREC; j++){
		i  = (((UI64)((rand()>>4) &	0xff))<< 0);
		i += (((UI64)((rand()>>4) &	0xff))<< 8);
		i += (((UI64)((rand()>>4) &	0xff))<<16);
		i += (((UI64)((rand()>>4) &	0xff))<<24);
		i += (((UI64)((rand()>>4) &	0xff))<<32);
		i += (((UI64)((rand()>>4) &	0xff))<<40);
		i += (((UI64)((rand()>>4) &	0xff))<<48);
		i += (((UI64)((rand()>>4) &	0xff))<<56);
		*(((PUI64)&vData[0])+j)	= i	;}

//		sort data

	dwTimeStart	= clock();

	MakeBind();

	SortBind();

	dwTimeStop = clock();
	cout <<	"Number of ticks	" << (dwTimeStop-dwTimeStart) << "\n";

#if	CMPFUNC
	cout <<	"Number	of compares	" << qwCmp << "\n";
#endif

//		output data

	fpDst =	fopen("dsth.bin", "wb");
	if(fpDst ==	NULL){
		goto exit0;}
	fwrite(pSort, 1, NUMREC*sizeof(UI64), fpDst);
	fclose(fpDst);

exit0:
	return(0);
}

/*----------------------------------------------------------------------*/
/*		MakeBind	seperate data in to	bins							*/
/*----------------------------------------------------------------------*/
static void	MakeBind(void)
{
static PUI64 pRec =	pData;				// ptr to org data
static PUI64 pEnd =	pData+NUMREC;
static PUI64 pTmp;
static UI64	 uRec;
static UI64	 i;

	for(i =	0; i < NUMBIN; i++){		// init	bin	counts
		pBinc[i] = 0;
	}

	pTmp = pBind;						// init	bin	ptrs
	for(i =	0; i < NUMBIN; i++){
		pBinp[i] = pTmp;
		pTmp +=	NUMREC;
	}

	while(pRec < pEnd){					// copy	data into bins
		uRec = *pRec++;
		i =	uRec>>SHFBIN;
		*(pBinp[i]++) =	uRec;
		pBinc[i] +=	1;
	}

	pTmp = pBind;						// init	bin	ptrs
	for(i =	0; i < NUMBIN; i++){
		pBinp[i] = pTmp;
		pTmp +=	NUMREC;
	}
}

/*----------------------------------------------------------------------*/
/*		SortBind	sort, xfer the bins									*/
/*----------------------------------------------------------------------*/
static void	SortBind(void)
{
PUI64 pOut;								// output data
UI32 uCnt;
UI64 i;

	pOut = pSort;						// init	pOut ptr

	for(i =	0; i < NUMBIN; i++){		// sort	the	bins
		uCnt = pBinc[i];
		if(uCnt){
			MakeGroups(pBinp[i], uCnt);
			MergeGroups(pBinp[i], pOut,	uCnt);
			if(pDone !=	pOut)
				memcpy(pOut, pBinp[i], uCnt*sizeof(UI64));
			pOut +=	uCnt;
		}
	}
}

#if	CMPFUNC
/*----------------------------------------------------------------------*/
/*		ElemCmp		element	compare										*/
/*----------------------------------------------------------------------*/
static bool	ElemCmp(UI64 elem1,	UI64 elem2)
{
	qwCmp += 1;							/* bump	# of compares */
	return elem1 < elem2;				/* return compare value	*/
}
#endif

/*----------------------------------------------------------------------*/
/*		MakeGroups														*/
/*----------------------------------------------------------------------*/
static void	MakeGroups(PUI64 pv, UI32 uc)
{
PUI64	pe;
UI64	qw0, qw1;

	pe = pv	+ (uc&0xfffffffeul);
	while(pv < pe){
		qw0	= pv[0];
		qw1	= pv[1];
#if	CMPFUNC
		if(pfCmp(qw1, qw0)){
#else
		if(qw1 < qw0){
#endif
			pv[0] =	qw1;
			pv[1] =	qw0;}
		pv += 2;}
}

/*----------------------------------------------------------------------*/
/*		MergeGroups														*/
/*----------------------------------------------------------------------*/
static void	MergeGroups(PUI64 pv0, PUI64 pv1, UI32 uc)
{
PUI64	pGet;							/* current pass	base ptrs */
PUI64	pPut;
PUI64	pGet0;							/* working ptrs	*/
PUI64	pGet1;
PUI64	pEnd;
PUI64	pEnd0;
PUI64	pEnd1;
PUI64	pPutx;
UI64	qwRec0;							/* working records */
UI64	qwRec1;
UI64	qwGrpCnt;						/* group size */

	pGet = pv0;							/* set up for first	pass */
	pPut = pv1;
	qwGrpCnt = 2;

/*		start of a pass	*/

	while(qwGrpCnt < uc){				/* while not done */
		pGet1 =	pGet;					/* update pass ptrs	*/
		pPutx =	pPut;
		pEnd  =	pGet+uc;

/*		merge two groups */

		while(1){
			pGet0 =	pGet1;				/* update group	ptrs */
			pEnd0 =	pGet0 +	qwGrpCnt;
			if(pEnd0 >=	pEnd)
				pEnd0 =	pEnd;
			pGet1 =	pEnd0;
			pEnd1 =	pGet1 +	qwGrpCnt;
			if(pEnd1 >=	pEnd)
				pEnd1 =	pEnd;
			if(pGet0 ==	pEnd0)			/* if no data, end pass	*/
				break;
			if(pGet1 ==	pEnd1){			/* if only group0, copy	it,	end	pass */
				while(pGet0	!= pEnd0)
					*pPutx++ = *pGet0++;
				break;}

/*		repeatedly copy	the	lower of two records */
/*		until both groups copied */

			qwRec0 = *pGet0++;			/* get 1st 2 records */
			qwRec1 = *pGet1++;
			while(1){
#if	CMPFUNC
				if(!pfCmp(qwRec1, qwRec0)){	/* if 0	<= 1, put rec0 */
#else
				if(!(qwRec1	< qwRec0)){	/* if 0	<= 1, put rec0 */
#endif
					*pPutx++ = qwRec0;
					if(pGet0 < pEnd0){	/*	 if	not	end	0, get next	rec0 */
						qwRec0 = *pGet0++;
						continue;}
					else{				/*	 else copy rest	of grp1	*/
						*pPutx++ = qwRec1;
						while(pGet1	!= pEnd1)
							*pPutx++ = *pGet1++;
						break;}}		/*	 end two group merge */
				else{					/* else	1 <	0, put rec1	*/
					*pPutx++ = qwRec1;
					if(pGet1 < pEnd1){	/*	 if	not	end	1, get next	rec1 */
						qwRec1 = *pGet1++;
						continue;}
					else{				/*	 else copy rest	of grp0	*/
						*pPutx++ = qwRec0;
						while(pGet0	!= pEnd0)
							*pPutx++ = *pGet0++;
						break;}}}}		/*	 end two group merge */
		pEnd = pPut;					/* swap	ptrs */
		pPut = pGet;
		pGet = pEnd;
		qwGrpCnt <<= 1;}				/* bump	group size */

	pDone =	pGet;						/* set ptr to sorted data */
}
