/*----------------------------------------------------------------------*/
/*      rsortv.cpp      test sort                                       */
/*                                                                      */
/*      Jeff Reid       13DEC24  10:00                                  */
/*----------------------------------------------------------------------*/
/*      abbreviations for name prefixes                                 */
/*                                                                      */
/*      a       array of ...                                            */
/*      b       unsigned char (byte)                                    */
/*      cb      count of bytes in ...                                   */
/*      fp      FILE pointer                                            */
/*      p       pointer to ...                                          */
/*      dw      unsigned int (dword)                                    */
/*----------------------------------------------------------------------*/

/*      get rid of warnings for fopen */
#define _CRT_SECURE_NO_WARNINGS 1

#include <iostream>
#include <vector>

#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#define NUMREC (1024*4096)              // number of records

typedef unsigned __int64    UI64;
typedef unsigned __int64 *  PUI64;

static clock_t dwTimeStart;             // clock values
static clock_t dwTimeStop;

static PUI64    pSort;                  // ptr to sorted data
static FILE *   fpDst;                  // file desc dest file

PUI64 RadixSort(PUI64, PUI64, size_t);

int main( )
{
using namespace std;

    vector <UI64> vData (NUMREC);
    vector <UI64> vTemp (NUMREC);
    UI64 i, j;

//      generate data

    for(j = 0; j < NUMREC; j++){
        i  = (((UI64)((rand()>>4) & 0xff))<< 0);
        i += (((UI64)((rand()>>4) & 0xff))<< 8);
        i += (((UI64)((rand()>>4) & 0xff))<<16);
        i += (((UI64)((rand()>>4) & 0xff))<<24);
        i += (((UI64)((rand()>>4) & 0xff))<<32);
        i += (((UI64)((rand()>>4) & 0xff))<<40);
        i += (((UI64)((rand()>>4) & 0xff))<<48);
        i += (((UI64)((rand()>>4) & 0xff))<<56);
        *(((PUI64)&vData[0])+j) = i ;}

//      sort data

    dwTimeStart = clock();

    pSort = RadixSort(&vData[0], &vTemp[0], NUMREC);

    dwTimeStop = clock();
    cout << "Number of ticks    " << (dwTimeStop-dwTimeStart) << "\n";

//      output data

    fpDst = fopen("dstr.bin", "wb");
    if(fpDst == NULL){
        goto exit0;}
    fwrite(pSort, 1, NUMREC*sizeof(UI64), fpDst);
    fclose(fpDst);

exit0:
    return(0);
}

PUI64 RadixSort(PUI64 pData, PUI64 pTemp, size_t count)
{
size_t mIndex[8][256] = {0};            // index matrix
PUI64 pDst, pSrc, pTmp;
size_t i,j,m,n;
UI64 u;

    for(i = 0; i < count; i++){         // generate histograms
        u = pData[i];
        for(j = 0; j < 8; j++){
            mIndex[j][(size_t)(u & 0xff)]++;
            u >>= 8;
        }       
    }
    for(j = 0; j < 8; j++){             // convert to indices
        n = 0;
        for(i = 0; i < 256; i++){
            m = mIndex[j][i];
            mIndex[j][i] = n;
            n += m;
        }       
    }

    pDst = pTemp;                       // radix sort
    pSrc = pData;
    for(j = 0; j < 8; j++){
        for(i = 0; i < count; i++){
            u = pSrc[i];
            m = (size_t)(u >> (j<<3)) & 0xff;
            pDst[mIndex[j][m]++] = u;
        }
        pTmp = pSrc;
        pSrc = pDst;
        pDst = pTmp;
    }

    return(pSrc);
}
